import turtle
import random
import os


#Configuration
num_cell_h = 9 #Number of cells horizontally
num_cell_v = 9 #Number of cells vertically
num_mines = 10

debugging = False
ct = turtle.Turtle()
ct.ht()
ct.pu()
ct.speed(0) 

cellsize = 30
margin_left = cellsize
margin_top = 3 * cellsize
margin_bottom = 2 * cellsize

f_width = cellsize * num_cell_h    #width of mine field
f_height = cellsize * num_cell_v  #height of mine field
w_width = f_width + 2 * margin_left #Width of window
w_height = f_height + margin_top + margin_bottom #Height of window

# Colors
backgroundcolor = 'ghost white'
cellcolor = 'gray90'
cellbordercolor = 'gray33'
minecolor = 'black'
flagpolecolor = 'blue'
flagcolor = 'black'
crosscolor = 'red'
numbercolors = ['gray','blue','green','red','purple','maroon','turqoise3','gray13','gray25']


# Game Variables
gameover = False
minefield = []
minemap = []
fieldstatus = []
flagged = []
timer = 0
timer_started = False
num_unflagged = num_mines
level = 'Beginner'
best_times = {'Beginner' : {'name' : 'No one', 'time' : 999}, \
              'Intermediate' : {'name' : 'No one', 'time' : 999}, \
              'Expert' : {'name' : 'No one', 'time' : 999}}
              
    

#Grid turtle()
g = turtle.Turtle()
g.ht()
g.speed(0)
g.shape('turtle')


#Drawing turtle
p = turtle.Turtle()

# Timer Turtle
tt = turtle.Turtle()
tt.pu()
tt.ht()

# Mine Counter Turtle
mt = turtle.Turtle()
mt.ht()
mt.pu()
mt.speed(0)

# Best Time Turtle

bt = turtle.Turtle()


def read_ini_file():
    ini_file = open('MinesweeperHighScore.txt','r')
    ini_data = ini_file.read().split('\n')
    for line in ini_data:
        if len(line) > 0:
            line_token = line.split(':')
            best_times[line_token[0]]['name'] = line_token[1]
            best_times[line_token[0]]['time'] = int(line_token[2])
    ini_file.close

def update_ini_file():
    bt_data = ''
    for level in best_times:
        bt_data = bt_data + level + ':' + best_times[level]['name'] + ':' + str(best_times[level]['time']) + "\n"
        ini_file = open('MinesweeperHighScore.txt','w')
        ini_file.truncate(0)
        ini_file.seek(0)
        ini_file.write(bt_data)
        ini_file.close()
    


# Setup window
w = turtle.Screen()
w.title('Minesweeper')
w.bgcolor(backgroundcolor)
w.tracer(0,0)

# Smiley Turtle
w.register_shape('smiley.gif')
w.register_shape('smiley1.gif')
w.register_shape('smiley2.gif')
st = turtle.Turtle()


def setup_window():
    w.setup(width = w_width, height = w_height)
    w.setworldcoordinates(0,0,w_width,w_height)

    
def setup_grid():
    g.ht()
    g.speed(0)
    g.pu()
    g.pensize(1)
    g.pencolor(cellbordercolor)
    g.clear()
    gx = margin_left
    g.seth(0)
    for h in range(num_cell_v + 1):
        gy = margin_bottom + h*cellsize
        g.goto(gx,gy)
        g.pd()
        g.fd(f_width)
        g.pu()
    gy = margin_bottom
    g.seth(90)
    for v in range(num_cell_h + 1):
        gx = margin_left + v*cellsize
        g.goto(gx,gy)
        g.pd()
        g.fd(f_height)
        g.pu()


def setup_minecounter():
    mt.ht()
    mt.pu()
    mt.color('red')
    x = margin_left
    y = margin_bottom + f_height
    mt.goto(x,y)
    mt.clear()
    mt.write('{:0>3}'.format(num_unflagged), False, 'left', ('Typewriter', int(1.3*cellsize), 'bold'))


def update_minecounter():
    mt.clear()
    mt.write('{:0>3}'.format(num_unflagged), False, 'left', ('Typewriter', int(1.3*cellsize), 'bold'))



def setup_timer():
    tt.speed(0)
    tt.color('red')
    x = margin_left + f_width
    y = margin_bottom + f_height 
    tt.goto(x,y)
    tt.clear()
    tt.write('{:0>3}'.format(0), False, 'right', ('Typewriter', int(1.3*cellsize), 'bold'))
    tt.ht()
    
def update_timer():
    global timer
    if timer_started:
        timer = timer + 1
        tt.clear()
        tt.write('{:0>3}'.format(timer), False, 'right', ('Typewriter', int(1.3*cellsize), 'bold'))
        w.update()
        w.ontimer(update_timer, 1000)

def setup_drawing():
    p.ht()
    p.speed(0)
    p.pensize(1)
    p.pu()
    g.ht()



def setup_smiley():
    st.ht()
    st.pu()
    st.speed(0)
    st.goto(margin_left + f_width/2, margin_bottom + f_height + cellsize)

def update_smiley(status = 0):
    if status == 1:
        st.shape('smiley1.gif')
    elif status == 2:
        st.shape('smiley2.gif')
    else:
        st.shape('smiley.gif')
    st.showturtle()
    w.update()


def setup_besttime():
    read_ini_file()
    bt.ht()
    bt.speed(0)
    bt.pu()
    bt.goto(margin_left, 0)
    bt_display = ''
    for level in best_times:
        bt_display = bt_display + level + ':'
        bt_display = bt_display + ' ' + best_times[level]['name'] + ' ' + str(best_times[level]['time']) + "\n"
    bt.write(bt_display, False, 'Left', ('Typewriter', int(0.35*cellsize), 'normal'))

def update_besttime():
    bt.clear()
    bt_display = ''
    for level in best_times:
        bt_display = bt_display + level + ':'
        bt_display = bt_display + ' ' + best_times[level]['name'] + ' ' + str(best_times[level]['time']) + "\n"
    bt.write(bt_display, False, 'Left', ('Typewriter', int(0.35*cellsize), 'normal'))
    update_ini_file()




# Initialize minefield variables
def init_map():
    for r in range(num_cell_v):
        minefield.append([])
        minemap.append([])
        fieldstatus.append([])
        flagged.append([])
        for c in range(num_cell_h):
            minefield[r].append(0)
            minemap[r].append(0)
            fieldstatus[r].append(0)
            flagged[r].append(0)

        
def generate_minefield():
    fields = random.sample(range(num_cell_h * num_cell_v), num_mines)
    for r in range(num_cell_v):
        for c in range(num_cell_h):
            if (r*num_cell_v + c) in fields:
                minefield[r][c] = 1
            else:
                minefield[r][c] = 0
    update_minemap()
            

def update_minemap():
    for r in range(num_cell_v):
        for c in range(num_cell_h):
            fieldstatus[r][c] = 0
            flagged[r][c] = 0
            if minefield[r][c] == 0:
                minemap[r][c] = count_mines(r,c)

def count_mines(r,c):
    mines = 0
    for mr in range(r-1, r+2):
        if mr >= 0 and mr < num_cell_v:
            for mc in range(c-1,c+2):
                if mc >= 0 and mc < num_cell_h and not (mr == r and mc == c):
                    mines = mines + minefield[mr][mc]
    return mines


def is_inside_field(x,y):
    return x > margin_left and x < margin_left + f_width and \
           y > margin_bottom and y < margin_bottom + f_height


def clicked(x,y):
    global timer_started
    if is_inside_field(x,y) and not gameover:
        r = int(num_cell_v - 1 - ((y - margin_bottom) // cellsize))
        c = int(x - margin_left) // cellsize
        if not timer_started:
            timer_started = True
            w.ontimer(update_timer, 1000)
        show_cell(r,c)
        

def right_clicked(x,y):
    global num_unflagged
    if is_inside_field(x,y) and not gameover:
        r = int(num_cell_v - 1 - ((y - margin_bottom) // cellsize))
        c = int((x - margin_left) // cellsize)
        if fieldstatus[r][c] == 0:
            draw_flag(r,c)
            fieldstatus[r][c] = 1
            flagged[r][c] = 1
            num_unflagged = num_unflagged - 1
        elif flagged[r][c] == 1:
            flagged[r][c] = 0
            fieldstatus[r][c]= 0
            close_cell(r,c)
            num_unflagged = num_unflagged + 1
        update_minecounter()
              
   

#Open the clicked cell
# and either show a bomb/mine or a number
def show_cell(r,c):
    global gameover, timer_started
    if not gameover:
        if fieldstatus[r][c] == 0:
            fieldstatus[r][c] = 1
            if minefield[r][c] == 1:
                draw_mine(r,c)
                show_all_mines()
                show_wrong_flags()
                gameover = True
                timer_started = False
                update_smiley(2)
            else:
                show_number(r,c)
        elif flagged[r][c] == 0 and minemap[r][c] > 0 and \
             minemap[r][c] == count_flags(r,c):
            show_neighbors(r,c)

                

        if not gameover and winning_protocol():
            timer_started = False
            gameover = True
            update_smiley(1)
            num_unflagged = 0
            update_minecounter()
            show_all_flags()
            if timer < int(best_times[level]['time']):
                bt_name = w.textinput('Best Time', 'Congrats! You have achieved a new best time! Please enter your name for the score board:')
                best_times[level]['name'] = bt_name
                best_times[level]['time'] = timer
                update_besttime()
                w.listen()
        
def show_all_flags():
    for r in range(num_cell_v):
        for c in range(num_cell_h):
            if minefield[r][c] == 1 and flagged[r][c] == 0:
                draw_flag(r,c)
    





            
def count_flags(r,c):
    flags = 0
    for mr in range(r-1, r+2):
        if mr >= 0 and mr < num_cell_v:
            for mc in range(c-1,c+2):
                if mc >= 0 and mc < num_cell_h and not (mr == r and mc == c) \
                   and flagged[mr][mc] == 1:
                    flags = flags + 1
    return flags



        
def winning_protocol():
    winning_protocol = True
    for r in range(num_cell_v):
        for c in range (num_cell_h):
            if minefield[r][c] == 0:
                winning_protocol = winning_protocol and fieldstatus[r][c] == 1
    return winning_protocol
    



  

def open_cell(r,c):
    x = margin_left + c*cellsize
    y = margin_bottom + (num_cell_v-r-1)*cellsize
    p.pu()
    p.goto(x,y)
    p.seth(0)
    p.color(cellbordercolor, cellcolor)
    p.pd()
    p.begin_fill()
    for l in range(4):
        p.fd(cellsize)
        p.left(90)
    p.end_fill()
    p.pu()

def close_cell(r,c):
    x = margin_left + c*cellsize
    y = margin_bottom + (num_cell_v-1-r)*cellsize
    p.goto(x,y)
    p.seth(0)
    p.color(cellbordercolor,backgroundcolor)
    p.pd()
    p.begin_fill()
    for l in range(4):
        p.fd(cellsize)
        p.left(90)
    p.end_fill()
    p.pu()
   

# Draw a flag at the cell r,c
def draw_flag(r,c):
    x = margin_left + c*cellsize + 0.1*cellsize
    y = margin_bottom + (num_cell_v-r-1)*cellsize + 0.1*cellsize
    p.pu()
    p.seth(0)
    p.goto(x,y)
    p.color(flagpolecolor,flagpolecolor)
    p.begin_fill()
    p.fd(0.6*cellsize)
    p.left(90)
    p.fd(0.1*cellsize)
    p.left(90)
    p.fd(0.1*cellsize)
    p.right(90)
    p.fd(0.1*cellsize)
    p.left(90)
    p.fd(0.15*cellsize)
    p.right(90)
    p.fd(0.4*cellsize)
    p.left(90)
    p.fd(0.1*cellsize)
    p.left(90)
    p.fd(0.4*cellsize)
    p.right(90)
    p.fd(0.15*cellsize)
    p.left(90)
    p.fd(0.1*cellsize)
    p.right(90)
    p.fd(0.1*cellsize)
    p.left(90)
    p.fd(0.1*cellsize)
    p.end_fill()
    x = x + 0.4*cellsize
    y = y + 0.35*cellsize
    p.goto(x,y)
    p.color(flagcolor, flagcolor)
    p.seth(90)
    p.begin_fill()
    p.fd(0.4*cellsize)
    p.left(120)
    p.fd(0.4*cellsize)
    p.left(120)
    p.fd(0.4*cellsize)
    p.end_fill()




def draw_mine(r,c):
    open_cell(r,c)
    x = margin_left + c*cellsize + 0.1*cellsize
    y = margin_bottom + (num_cell_v-r-1)*cellsize + 0.5*cellsize
    p.pu()
    p.goto(x,y)
    p.color(minecolor)
    p.pd()
    p.begin_fill()
    p.seth(0)
    # Start drawing a mine
    for s in range(9):
        p.forward(0.8*cellsize)
        p.left(160)
    p.end_fill()
    p.pu()


def draw_cross(r,c):
    x = margin_left + c*cellsize + 0.1*cellsize
    y = margin_bottom + (num_cell_v-1-r)*cellsize + 0.1*cellsize
    p.seth(0)
    p.pensize(2.5)
    p.goto(x,y)
    p.color(crosscolor)
    p.seth(45)
    p.pd()
    p.goto(x + 0.8*cellsize, y + 0.8*cellsize)
    p.pu()
    p.goto(x, y+0.8*cellsize)
    p.seth(-45)
    p.pd()
    p.goto(x + 0.8*cellsize, y)
    p.pu()
    p.pensize(1)


# show a number at the cell r,c(the number is taken from mine map)
def show_number(r,c):
    open_cell(r,c)
    if minemap[r][c] > 0:
        p.color(numbercolors[minemap[r][c]])
        x = margin_left + c*cellsize + 0.5*cellsize
        y = margin_bottom + (num_cell_v-r-1)*cellsize
        p.pu()
        p.goto(x,y)
        p.write(minemap[r][c],False,"center", ('Typewriter', int(0.7*cellsize), 'bold'))
    else:
        show_neighbors(r,c)

def show_neighbors(r,c):
    for mr in range(r-1,r+2):
        if mr >= 0 and mr < num_cell_v:
            for mc in range(c-1,c+2):
                if mc >= 0 and mc < num_cell_h \
                   and not (mr== r and mc == c) \
                   and fieldstatus[mr][mc] == 0:
                    show_cell(mr,mc)


def show_all_mines():
     for r in range(num_cell_v):
        for c in range(num_cell_h):
            if minefield[r][c] == 1 and fieldstatus[r][c] == 0:
                draw_mine(r,c)

def show_wrong_flags():
    for r in range(num_cell_v):
        for c in range(num_cell_h):
            if minefield[r][c] == 0 and flagged[r][c] == 1:
                draw_mine(r,c)
                draw_cross(r,c)

def expert():
    global level
    level = "Expert"
    set_level()

def intermediate():
    global level
    level = "Intermediate"
    set_level()

def beginner():
    global level
    level = "Beginner"
    set_level()



def set_level():
    global num_cell_v, num_cell_h, num_mines, f_width, f_height, w_width, w_height
    if level == "Beginner":
        num_cell_v = 9
        num_cell_h = 9
        num_mines = 10
    elif level == 'Intermediate':
        num_cell_v = 16
        num_cell_h = 16
        num_mines = 40
    elif level == 'Expert':
        num_cell_v = 16
        num_cell_h = 30
        num_mines = 99
    
    f_width = cellsize * num_cell_h    #width of mine field
    f_height = cellsize * num_cell_v  #height of mine field
    w_width = f_width + 2 * margin_left #Width of window
    w_height = f_height + margin_top + margin_bottom #Height of window

    minesweeper()
    
    

def start(x=0,y=0):
    global gameover
    global timer_started
    global timer
    global num_unflagged

    gameover = False
    num_unflagged = num_mines
    timer_started = False
    timer = 0
    p.clear()
    generate_minefield()
    setup_smiley()
    update_smiley()
    setup_timer()
    setup_minecounter()
    update_minecounter()
    
    if debugging:
        show_cheats()

        
def show_cheats():
    print('Minefield = ')
    for r in range(num_cell_v):
        print(minefield[r])
    print()
    print('Minemap = ')
    for r in range(num_cell_v):
        print(minemap[r])
    print()
    

def minesweeper(init = False):  
    if init:
        beginner()
    else:
        setup_window()
        setup_grid()
        init_map()
        setup_drawing() 
        setup_besttime()
        start()
        w.update()
        w.listen()
        w.onclick(clicked)
        w.onclick(right_clicked, btn=2)
        w.onkey(show_cheats,'c')
        w.onkey(expert,'e')
        w.onkey(intermediate,'i')
        w.onkey(beginner,'b')
        st.onclick(start)
        w.mainloop()


minesweeper()


#Y          Y   EEEEEEEEEE   EEEEEEEEEE   TTTTTTTTTTT
# Y        Y    E            E                 T             
#  Y      Y     E            E                 T  
#   Y    Y      E            E                 T
#    YYYY       EEEEEEEEEE   EEEEEEEEEE        T
#     YY        E            E                 T
#     YY        E            E                 T
#     YY        E            E                 T
#     YY        EEEEEEEEEE   EEEEEEEEEE        T
     

